#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <elf.h>
#include <fcntl.h>
#include <ncurses.h>
#include "gui.h"
#include "menu.h"
#include "main.h"

/* Global variables */
Browser browser;
int fd = -1;

/* Pointers to menu commands */
#define COMMANDS 6
const MenuItem cmd[COMMANDS] = 
{
     { "exit", "exit to console", cmd_quit },
     { "quit", "exit to console", cmd_quit },
     { "help", "display help"   , cmd_help },
     { "open", "open core image", cmd_open },
     { "seek", "search the stack for a string, return its memory address(s)", cmd_seek },
     { "word", "search the stack for a 32bit word", cmd_word }
};

void cmd_quit(char *_)
{
   refresh();
   UI_End();
   exit(1);
}

void cmd_help(char *_)
{
   int i;
   
   UI_AddBrowser(&browser, "Available commands:");
   for (i=0;i<COMMANDS;i++)
     UI_AddBrowser(&browser, "%s    %s",cmd[i].command,cmd[i].description);
   UI_AddBrowser(&browser, "-- eof --");
   UI_AddBrowser(&browser, " ");
}

void cmd_open(char *fname)
{
   if (fname == NULL) {
      UI_AddBrowser(&browser, "Usage: open <core dump path>");
      return;
   }
   
   // close file descriptor in case something else is already open
   close(fd);
   
   // attempt accessing file
   if ((fd = open(fname+0x1, O_RDONLY))<0) {
      UI_AddBrowser(&browser, "open: unable to access %s",fname);
      return;
   }
   
   // read Elf binary header and check binary type
   if (Segment_init(fd) < 0) {
      UI_AddBrowser(&browser, "open: not a core dump, closing file descriptor.");
      close(fd);
      fd = -1;
      return;
   }
   
   UI_AddBrowser(&browser, "open: %s (core dump) loaded.",fname);
}

void cmd_seek(char *word)
{  
   /* Error checking */
   if (word == NULL) 
     { UI_AddBrowser(&browser, "Usage: seek <string to seek>"); return; }
   if (fd < 0) { UI_AddBrowser(&browser, "seek: no file open"); return; }
   
   // initialize segment again, just to be sure we are at the top
   // of the segment headers table.
   Segment_init(fd);
   
   for (;;) {
      Elf32_Phdr Header;
      unsigned long int pos, i;
      
      // read a segment header from the headers table
      if (read(fd, &Header, sizeof(Elf32_Phdr))<0) {
	 UI_AddBrowser(&browser, "seek: read() error, aborting.");
	 break;
      }
      
      // if the segment type is invalid, that most likely means we reached
      // the end of the headers table.
      if (Segment_valid(Header.p_type)<0)
	break;
      
      // the offset for the next segment header will be saved in pos
      pos = lseek(fd, 0, SEEK_CUR);
      
	{
	   const int size = Header.p_filesz;
	   char buffer[size];
	   
	   memset(&buffer, 0, sizeof(buffer));
	   
	   // jump to the segment's offset in the file
	   if (lseek(fd, Header.p_offset, SEEK_SET) != Header.p_offset) {
	      UI_AddBrowser(&browser, "seek: unable to lseek to segment content position (0x%x), aborting.",Header.p_offset);
	      break;
	   }
	   
	   // read the segment's content into buffer
	   if (read(fd, &buffer, size) != size)
	     break;
	   
	   // clear all zero characters so we could strstr() in peace
	   for (i=0;i<size;i++)
	     if (buffer[i] == 0) buffer[i] = 1; // whatever ..
	   buffer[size-1] = 0;
	   
	   // look for our string (finally ... :)
	   if (strstr(buffer, word+1)) {
	      char *ptr;
	      
	      ptr = strstr(buffer, word+1);
	      ptr -= buffer;
	      
	      UI_AddBrowser(&browser, "* Query found on 0x%x (%s)",
			    ptr+Header.p_vaddr,Segment_type(Header.p_type));
	   }
	   
	   // return to next segment header position
	   if (lseek(fd, pos, SEEK_SET) != pos) {
	      UI_AddBrowser(&browser, "seek: unable to return to next segment header position, aborting.");
	      break;
	   }
	}
   }
   UI_AddBrowser(&browser, "Stack search for string \"%s\" completed.",word+1);
}

void cmd_word(char *word)
{  
   int address;
   
   /* Error checking */
   if (word == NULL) 
     { UI_AddBrowser(&browser, "Usage: word <32bit word in hexadecimal>"); return; }
   if (fd < 0) { UI_AddBrowser(&browser, "word: no file open"); return; }
   
   sscanf(word, " %x",&address);
   
   // initialize segment again, just to be sure we are at the top
   // of the segment headers table.
   Segment_init(fd);
   
   for (;;) {
      Elf32_Phdr Header;
      unsigned long int pos, i;
      
      // read a segment header from the headers table
      if (read(fd, &Header, sizeof(Elf32_Phdr))<0) {
	 UI_AddBrowser(&browser, "word: read() error, aborting.");
	 break;
      }
      
      // if the segment type is invalid, that most likely means we reached
      // the end of the headers table.
      if (Segment_valid(Header.p_type)<0)
	break;
      
      // the offset for the next segment header will be saved in pos
      pos = lseek(fd, 0, SEEK_CUR);
      
      for (i=0;i<Header.p_filesz/4;i++)
	{
	   char word[4];
	   int Offset, w;
	   
	   // get the word's offset in the file
	   Offset = Header.p_offset + i*4;
	   
	   if (lseek(fd, Offset, SEEK_SET) != Offset) {
	      UI_AddBrowser(&browser, "word: unable to jump to word offset 0x%x",Offset);
	      return;
	   }
	   if (read(fd, &word, 4)<0) {
	      UI_AddBrowser(&browser, "word: read() error - unable to read word on 0x%x",Offset);
	      return;
	   }
	   memcpy(&w, &word, 4);
	   if (w == address) {
	      UI_AddBrowser(&browser, "* word found on 0x%x [file offset 0x%x] (%s)",
			    Header.p_vaddr+i*4,Offset, Segment_type(Header.p_type));
	   }
	}
      
      if (lseek(fd, pos, SEEK_SET) != pos) {
	 UI_AddBrowser(&browser, "word: unable to return to next segment header position, aborting.");
	 break;
      }
   }
   UI_AddBrowser(&browser, "Stack search for 32bit word 0x%x completed.",address);
}


int main(int argc, char *argv[])
{
   UI_Init();
   
   UI_InitBrowser(&browser,0,0,COLS,LINES-1,1);
   UI_AddBrowser(&browser, "Segment debugger version %s (c) Copyright 2000 sectorx",VERSION);
   UI_AddBrowser(&browser, "http://xorteam.cjb.net");
   UI_RenderBrowser(&browser);
   
   do {
      char *ptr;
      int i;
      
      ptr = UI_Input(0,LINES-1,COLS,2);
      for (i=0;i<COMMANDS;i++)
	if (!strncmp(cmd[i].command, ptr, strlen(cmd[i].command))) {
	   char *parameter;
	   
	   parameter = strstr(ptr, " ");
	   cmd[i].func(parameter);
	}
      if (!strcmp(ptr, "")) UI_AddBrowser(&browser, " ");
      
      UI_RenderBrowser(&browser);
   } while (1);
}
